from functools import wraps

def real_time_decoding(func):
    """A decorator for real-time decoding"""
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        #call the original function
        outputs = func(self, *args, **kwargs)

        #deocde the outputs token by token
        for output in outputs:
            output_tokens = output[0][-1]
            print(self.tokenizer.decode(output_tokens))

        return outputs
    return wrapper



from types import NoneType
from typing import Generator, Optional, Sequence, Union


from ctransformers import AutoModelForCausalLM, AutoTokenizer
from ctransformers.hub import AutoConfig
from ctransformers.llm import Config


class CInference:
    def __init__(
        self,
        model_path: str = None,
        model_type: Optional[str] = None,
        model_file: Optional[str] = None,
        config: Optional[Union[AutoConfig, Config]] = None,
        lib: Optional[str] = None,
        local_files_only: bool = False,
        revision: Optional[str] = None,
        hf: bool = False,
        **kwargs
    ):
        self.model = AutoModelForCausalLM.from_pretrained(
            model_path,
            model_type=model_type,
            model_file=model_file,
            config=config,
            lib=lib,
            local_files_only=local_files_only,
            revision=revision,
            hf=hf,
            **kwargs
        )
        if hf:
            self.tokenizer = AutoTokenizer.from_pretrained(model_path)
        else:
            self.tokenizer = None
    def run(
            self, 
            prompt: str,
            max_new_tokens: Optional[int] = None,
            top_k: Optional[int] = None,
            top_p: Optional[float] = None,
            temperature: Optional[float] = None,
            repetition_penalty: Optional[float] = None,
            last_n_tokens: Optional[int] = None,
            seed: Optional[int] = None,
            batch_size: Optional[int] = None,
            threads: Optional[int] = None,
            stop: Optional[Sequence[str]] = None,
            stream: Optional[bool] = True,
            reset: Optional[bool] = None
        ) -> Union[str, Generator[str, NoneType, NoneType]]:
        if self.tokenizer:
            return self.model.__call__(
                prompt,
                max_new_tokens=max_new_tokens,
                top_k=top_k,
                top_p=top_p,
                temperature=temperature,
                repetition_penalty=repetition_penalty,
                last_n_tokens=last_n_tokens,
                seed=seed,
                batch_size=batch_size,
                threads=threads,
                stop=stop,
                stream=stream,
                reset=reset
            )
        else:
            tokens = self.model.tokenize(prompt)
            return self.model.generate(
                tokens,
                top_k=top_k,
                top_p=top_p,
                temperature=temperature,
                repetition_penalty=repetition_penalty,
                last_n_tokens=last_n_tokens,
                seed=seed,
                batch_size=batch_size,
                threads=threads,
                reset=reset
            )


from exa.exa import exa

print(exa)

#inference
from exa.inference.ctransformers import CInference
from exa.inference.gptq import GPTQInference
from exa.inference.hf import Inference
from exa.inference.kosmos import Kosmos
from exa.inference.mmi import MultiModalInference
from exa.inference.text_to_video import TextToVideo

#quant
from exa.quant.main import Quantize

#inference
from exa.utils.decoding_wrapper import real_time_decoding
from exa.utils.deploy import Deploy

#utils
from exa.utils.metric_logger import Logging



from exa.inference.gptq import GPTQInference
from exa.inference.hf import Inference
from exa.quant.main import Quantize

from exa.inference.text_to_video import TextToVideo
from exa.inference.mmi import MultiModalInference




import uvicorn
from fastapi import FastAPI, HTTPException
from pydantic import BaseModel
from slowapi import _rate_limit_exceeded_handler
from slowapi.errors import RateLimitExceeded
from slowapi.middleware import SlowAPIMiddleware
from slowapi.util import get_remote_address

app = FastAPI()
app.add_exception_handler(RateLimitExceeded, _rate_limit_exceeded_handler)
app.add_middleware(SlowAPIMiddleware)

class Prompt(BaseModel):
    text: str
    max_length: int = 20

class Deploy:
    def __init__(
        self,
        llm,
        app: FastAPI = None,
        host: str = "0.0.0.0",
        port: int = 8000,
    ):
        self.llm = llm
        self.app = app if app else FastAPI()
        self.host = host
        self.port = port

        @self.app.on_event("startup")
        def load_model():
            self.load_model()

        @self.app.post("/generate/")
        async def generate(prompt: Prompt):
            return self.generate(prompt)
        
    def load_model(self):
        global inference
        inference = self.llm

    async def generate(self, prompt: Prompt):
        try:
            return {"generated_text": inference.run(prompt.text)}
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))

    def run(self):
        uvicorn.run(self.app, host=self.host, port=self.port)

# if __name__ == "__main__":
#     llm = Inference(model_id="gpt2-small")  # replace with your LLM
#     deploy = Deploy(llm)
#     deploy.run()